print("import: 'libcst'")
import libcst

print("import: 'libcst._nodes'")
import libcst._nodes

print("import: 'libcst._nodes.tests'")
import libcst._nodes.tests

print("import: 'libcst._parser'")
import libcst._parser

print("import: 'libcst._parser.conversions'")
import libcst._parser.conversions

print("import: 'libcst._parser.parso'")
import libcst._parser.parso

print("import: 'libcst._parser.parso.pgen2'")
import libcst._parser.parso.pgen2

print("import: 'libcst._parser.parso.python'")
import libcst._parser.parso.python

print("import: 'libcst._parser.parso.tests'")
import libcst._parser.parso.tests

print("import: 'libcst._parser.tests'")
import libcst._parser.tests

print("import: 'libcst._parser.types'")
import libcst._parser.types

print("import: 'libcst._parser.types.tests'")
import libcst._parser.types.tests

print("import: 'libcst.codegen'")
import libcst.codegen

print("import: 'libcst.codegen.tests'")
import libcst.codegen.tests

print("import: 'libcst.codemod'")
import libcst.codemod

print("import: 'libcst.codemod.commands'")
import libcst.codemod.commands

print("import: 'libcst.codemod.commands.tests'")
import libcst.codemod.commands.tests

print("import: 'libcst.codemod.tests'")
import libcst.codemod.tests

print("import: 'libcst.codemod.visitors'")
import libcst.codemod.visitors

print("import: 'libcst.codemod.visitors.tests'")
import libcst.codemod.visitors.tests

print("import: 'libcst.helpers'")
import libcst.helpers

print("import: 'libcst.helpers.tests'")
import libcst.helpers.tests

print("import: 'libcst.matchers'")
import libcst.matchers

print("import: 'libcst.matchers.tests'")
import libcst.matchers.tests

print("import: 'libcst.metadata'")
import libcst.metadata

print("import: 'libcst.metadata.tests'")
import libcst.metadata.tests

print("import: 'libcst.testing'")
import libcst.testing

print("import: 'libcst.tests'")
import libcst.tests

